"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
var lodash_1 = require("lodash");
var inflection_1 = require("inflection");
var url = require("url");
var querystring = require("querystring");
var env = require("./env");
function swaggerToDataSource(swagger, https, runtimePath, endPoint) {
    var scheme = https
        ? 'https'
        : swagger.schemes
            ? swagger.schemes[0]
            : 'http';
    var basePath;
    if (!runtimePath) {
        basePath =
            endPoint ||
                (swagger.host
                    ? scheme + "://" + swagger.host + (swagger.basePath || '')
                    : swagger.basePath);
    }
    else {
        basePath = runtimePath;
    }
    if (basePath) {
        basePath = basePath.replace(/([^:]\/)\/+/g, '$1');
    }
    var auth = {
        type: 'none',
        clientId: null,
        url: null,
        name: null,
        value: null,
        scopes: [],
        params: []
    };
    if (swagger.securityDefinitions) {
        var oauthSecurity_1 = lodash_1.find(swagger.securityDefinitions, function (def, defName) {
            return def.type == 'oauth' || def.type == 'oauth2';
        });
        if (oauthSecurity_1) {
            auth.type = 'oauth';
            if (oauthSecurity_1.authorizationUrl) {
                var uri = url.parse(oauthSecurity_1.authorizationUrl, true);
                if (uri) {
                    var protocol = uri.protocol, host = uri.host, pathname = uri.pathname, query = uri.query;
                    if (!('scopes' in query) && oauthSecurity_1.scopes) {
                        query.scope = Object.keys(oauthSecurity_1.scopes).join(' ');
                    }
                    if (!('response_type' in query)) {
                        query.response_type = 'token';
                    }
                    auth.url = protocol + "//" + host + pathname + "?" + querystring.stringify(query);
                }
                else {
                    auth.url = oauthSecurity_1.authorizationUrl;
                }
            }
            else {
                ['tokenEndpoint', 'authEndpoint', 'clientId', 'clientSecret'].forEach(function (key) {
                    if (oauthSecurity_1[key]) {
                        auth[key] = oauthSecurity_1[key];
                    }
                });
                if (oauthSecurity_1.scopes) {
                    auth.scopes = Object.keys(oauthSecurity_1.scopes).join(' ');
                }
            }
        }
        var azureadSecurity_1 = lodash_1.find(swagger.securityDefinitions, function (def, defName) {
            return def.type == 'azuread';
        });
        if (azureadSecurity_1) {
            auth.type = 'azuread';
            if (azureadSecurity_1.authorizationUrl) {
                var uri = url.parse(azureadSecurity_1.authorizationUrl, true);
                if (uri) {
                    var protocol = uri.protocol, host = uri.host, pathname = uri.pathname, query = uri.query;
                    if (!('scopes' in query) && azureadSecurity_1.scopes) {
                        query.scope = lodash_1.uniq(Object.keys(azureadSecurity_1.scopes).concat('openid', 'profile')).join(' ');
                    }
                    auth.url = protocol + "//" + host + pathname + "?" + querystring.stringify(query);
                }
                else {
                    auth.url = azureadSecurity_1.authorizationUrl;
                }
            }
            else {
                ['tokenEndpoint', 'authEndpoint', 'clientId', 'clientSecret'].forEach(function (key) {
                    if (azureadSecurity_1[key]) {
                        auth[key] = azureadSecurity_1[key];
                    }
                });
                if (azureadSecurity_1.scopes) {
                    auth.scopes = Object.keys(azureadSecurity_1.scopes).join(' ');
                }
            }
        }
        var apiKey = lodash_1.find(swagger.securityDefinitions, function (def, defName) {
            return def.type == 'apiKey';
        });
        var apiKeyValue = Object.keys(swagger.securityDefinitions).find(function (i) { return swagger.securityDefinitions[i].type == 'apiKey'; });
        if (apiKey && apiKeyValue) {
            auth.type = 'apiKey';
            auth.name = apiKey.name;
            auth.in = apiKey.in;
            auth.value = apiKeyValue;
        }
        var basicAuth = lodash_1.find(swagger.securityDefinitions, function (def, defName) {
            return def.type == 'basic';
        });
        if (basicAuth) {
            auth.type = 'basic';
        }
    }
    function parseProperties(definition) {
        return lodash_1.map(definition.properties, function (property, name) {
            var result = Object.assign({ name: name }, property);
            if (property.$ref) {
                result.type = property.$ref.replace('#/definitions/', '');
                result.ref = true;
            }
            else if (property.type == 'array') {
                result.array = true;
                if (property.items) {
                    if (property.items.$ref) {
                        result.type = property.items.$ref.replace('#/definitions/', '');
                    }
                    else if (property.items.type) {
                        result.type = property.items.type;
                    }
                }
            }
            else if (property.properties) {
                result.properties = parseProperties(property);
            }
            result.key = !!property['x-key'];
            result.originalName = property['x-original-name'];
            result.navigation = property['x-navigation'];
            result.foreignKey = property['x-foreignKey'];
            result.required =
                definition.required && definition.required.indexOf(result.name) != -1;
            result.nullable = property['x-nullable'] != false && !result.required;
            result.$ref = property.$ref;
            return result;
        });
    }
    function parseSchema(definition, name) {
        var properties = parseProperties(definition);
        if (swagger.info['x-connector']) {
            properties = lodash_1.filter(properties, function (p, n) {
                return !p.array;
            });
        }
        return {
            name: name,
            resource: definition['resource'],
            table: definition['x-table'],
            view: definition['x-view'],
            parameters: definition['x-parameters'],
            entitySet: inflection_1.pluralize(name),
            properties: properties,
            required: definition.required
        };
    }
    function parseEnum(e, name) {
        return {
            name: name,
            flags: e['x-flags'],
            members: e['enum'].map(function (i) {
                return {
                    text: i.text,
                    value: i.value
                };
            }),
        };
    }
    var enums = swagger.components && swagger.components.schemas ? lodash_1.map(swagger.components.schemas, parseEnum) : [];
    var schemas = lodash_1.map(swagger.definitions, parseSchema);
    function responseSchema(path, schema) {
        var ref = null;
        if (!schema) {
            return null;
        }
        if (schema.type == 'array') {
            ref = schema.items.$ref;
        }
        else if (schema.$ref) {
            ref = schema.$ref;
        }
        var name = null;
        if (ref) {
            name = ref.replace('#/definitions/', '');
        }
        return name;
    }
    var operations = lodash_1.flatten(lodash_1.map(swagger.paths, function (path, name) {
        return lodash_1.map(path, function (operation, method) {
            var _a = operation.parameters, parameters = _a === void 0 ? [] : _a;
            var url = parameters
                .filter(function (param) { return param.in == 'path'; })
                .reduce(function (url, param) {
                return url.split('/').map(function (s) {
                    if (s.includes(',')) {
                        return s.replace(new RegExp("{" + param.name + "}", 'g'), param.type == 'string' && !param.format
                            ? param.name + "={HttpUtility.UrlEncode(" + env.variable(param.name) + ")}"
                            : param.name + "={" + env.variable(param.name) + "}");
                    }
                    else {
                        return s.replace(new RegExp("{" + param.name + "}", 'g'), param.type == 'string' && !param.format
                            ? "{HttpUtility.UrlEncode(" + env.variable(param.name) + ")}"
                            : "{" + env.variable(param.name) + "}");
                    }
                }).join('/');
            }, name);
            return {
                path: url,
                method: method,
                name: operation.operationId,
                parameters: lodash_1.map(parameters, function (parameter) {
                    var type = parameter.schema
                        ? responseSchema(name, parameter.schema)
                        : parameter.type;
                    var parameterName = parameter.name;
                    if (!parameterName) {
                        if (parameter.in == 'body') {
                            parameterName = env.variable(env.unwrapType(type));
                        }
                        else {
                            throw new Error('Parameter name not specified.');
                        }
                    }
                    return {
                        in: parameter.in,
                        name: parameterName,
                        required: parameter.required,
                        default: parameter.default,
                        items: parameter.items,
                        type: type,
                        format: parameter.format
                    };
                }),
                responses: lodash_1.map(operation.responses, function (response, status) {
                    return {
                        status: status,
                        schema: response.schema
                    };
                })
            };
        });
    }));
    var resources = lodash_1.groupBy(operations, function (_a) {
        var path = _a.path;
        return path.replace(/\//g, '').replace(/\(?'?\${[^}]*}'?\)?,?/g, '');
    });
    return {
        name: swagger.info ? swagger.info.title : '',
        authorization: auth.type != 'none' ? auth : null,
        connector: swagger.info['x-connector'],
        connectorVersion: swagger.info['x-connector-version'],
        type: swagger.info['x-type'],
        version: swagger.info['x-odata-version'],
        withCredentials: swagger.info['x-withCredentials'],
        concurrency: swagger.info['x-concurrency'],
        cascadeDelete: swagger.info['x-cascade-delete'] != false,
        basePath: basePath,
        resources: resources,
        schemas: schemas,
        enums: enums,
        host: runtimePath || swagger.host
    };
}
exports.swaggerToDataSource = swaggerToDataSource;
//# sourceMappingURL=DataSource.js.map